from encoders.helpers import *
from extractors.strategy import *

from extractors.solvers.msat import *
from checkers import *

from lazy import *

import common



class DTPUStrategyExtractor(object):

    def __init__(self, quantelim, environment):
        self.quantelim = quantelim
        self.environment = environment

        self.solver = MathSAT5Solver(self.environment)
        self.checker = SyntacticChecker()
        return


    def extractStrategy(self, clausifier):
        st = self.environment.symbolTable

        yvars = clausifier.yvars
        xvars = clausifier.xvars

        print("Starting strategy extraction...")

        print("  Declaring Variables...")
        coeff = []
        for nv in yvars:
            coeff.append(nv)
            self.solver.declareVariable(nv)

        for x in xvars:
            self.solver.declareVariable(x)

        print("  Declaring Atoms...")
        ba2atom = {}
        bavars = []
        for a in clausifier.atoms:
            nv = st.createNewFreshSymbol(Types.BOOL, "_BA_%d")
            self.solver.declareVariable(nv)
            ba2atom[nv.name] = a
            bavars.append(nv)
            self.solver.addAssertion(Iff(a, nv))

        self.solver.addAssertion(clausifier.getPsi().simplify())
        self.solver.addAssertion(clausifier.getGamma().simplify())

        strategy = [None]

        tmp = MathSAT5Solver(Environment())
        for y in yvars:
            tmp.declareVariable(y)
        tmp.addAssertion(clausifier.getGamma())

        print("  All SAT...")
        self.solver.allSat(bavars,
                           lambda model: self.callback(model,
                                                       strategy,
                                                       clausifier,
                                                       ba2atom,
                                                       tmp))


        wc = not tmp.solve()
        tmp.destroy()

        if not wc:
            return None
        return strategy[0]


    def callback(self, model, strategy, clausifier, ba2atom, completionChecker):
        if strategy[0] is None:
            strategy[0] = PiecewiseStrategy(clausifier.yvars)
        s = strategy[0]

        polyconj = []
        for m in model:
            (var, negated) = self.solver.term2VarName(m)
            atom = ba2atom[var]
            if negated:
                polyconj.append(Not(atom))
            else:
                polyconj.append(atom)

        polyconj.append(clausifier.getGamma())
        polyhedron = And(polyconj)

        shadow = Exists(clausifier.xvars, polyhedron)
        elim = self.quantelim.removeQuantifiers(shadow)
        shadow = elim.simplify()

        # expandStrategy
        ex = LazyStrategyExtractor(Environment())
        strat = ex.extractStrategy(clausifier, gamma=shadow, psi=polyhedron)
        s.addPiece(shadow, strat)

        completionChecker.addAssertion(Not(shadow))
        return 1 if completionChecker.solve() else 0
