from encoders.helpers import *
from extractors.strategy import *

from extractors.solvers.msat import *
from checkers import *

import common
import random


class OneShotLinearWeakStrategyExtractor(object):

    def __init__(self, environment):
        self.environment = environment
        self.solver = MathSAT5Solver(environment)
        return


    def extractStrategy(self, clausifier):
        st = self.environment.symbolTable
        
        yvars = clausifier.yvars
        xvars = clausifier.xvars

        s = LinearStrategy(yvars)

        print("Starting strategy extraction...")

        print("  Declaring Variables...")
        coeff = []
        for _ in range(0, len(xvars)):
            row = []
            for _ in range(0, len(yvars) + 1):
                nv = st.createNewFreshSymbol(Types.REAL, "_C_%d")
                self.solver.declareVariable(nv)
                row.append(nv)
            coeff.append(row)

        print("  Encoding problem as SMT formula...")
        formula = self.encode(coeff, clausifier)
    
        print("  Asserting SMT formula...")
        self.solver.addAssertion(formula)
    
        print("  Solving SMT formula...")
        if not self.solver.solve():
            common.error("No linear strategy exists!")
            exit(1)
                
        print("  Extracting model...")
        for i in range(0, len(xvars)):
            row = coeff[i]
            x = xvars[i]

            vals = []
            for c in row:
                vals.append(self.solver.getModelValue(c))
            s.addPolicy(x, vals)
    
        print("Done.")
        return s


    def encode(self, coeff, clausifier):
        xvars = clausifier.xvars
        yvars = clausifier.yvars
        ycount = len(yvars)

        print("      Preparing formula definitions...")
        formula = clausifier.getPsi()


        print("      Building problem formula...")
        for i in range(0, len(xvars)):
            row = coeff[i]
            x = xvars[i]

            f = row[-1]
            for j in range(0, len(row) - 1):
                c = row[j]
                f = Plus(Times(c, yvars[j]), f)

            formula = Definition(x.name, f, formula)

        formula = formula.simplify()
        resList = self.allCases(clausifier, list(yvars), formula)
        return And(resList)


    def allCases(self, clausifier, choices, base, count=0, total=None):
        if len(choices) == 0:
            return [base]

        pivot = choices.pop()
        (l,u) = clausifier.getYBound(pivot)

        if total is None:
            total = len(choices)
        count = count+1
        
        sonList = self.allCases(clausifier, choices, base, count, total)

        perc = 100 / (2**count)

        print("        [%d/%d]\tGenerating cases for variable %s.\t\t(%.2f%% overall completion)" % (total-count+2, total+1, pivot.name, perc))
        res = []
        for s in sonList:
            res.append(Definition(pivot.name, Constant(l), s))
            res.append(Definition(pivot.name, Constant(u), s))

        return res

    pass
