import os
import sys
import re
import random
 
from subprocess import Popen, PIPE


import common

from encoders.printers import SmtPrinter
from formula import *

 

class Z3QuantifierElimination(object):
    
    def __init__(self, z3_path):
        self.z3 = z3_path
        return

    
    def removeQuantifiers(self, formula):
        f = common.getIOBuffer()

        f.write(common.toBytes("(set-option :print-success false)\n"))
        f.write(common.toBytes("(set-logic LRA)\n"))
        f.write(common.toBytes("\n"))
        
        for v in formula.getDependencies():
            f.write(common.toBytes("(declare-fun %s %s)\n" % (v.name, v.typename.toString())))

        f.write(common.toBytes("\n"))
        f.write(common.toBytes("(simplify\n"))

        p = SmtPrinter(f, strictSyntax=True, useBytes=True, smt2version=True)
        p.printer(formula)

        # import sys
        # print("--QE------------------")
        # p = SmtPrinter(sys.stdout, strictSyntax=True, useBytes=False, smt2version=True)
        # p.printer(formula)
        # print("\n----------------------\n")

        f.write(common.toBytes(")\n"))

        data = f.getvalue()
        f.close()

        output = self.callZ3(data)
        output = common.convertString(output)
        res = output.replace("~", "-");
        
        # print("--QE Result-----------")
        # print(res)
        # print("----------------------\n")

        return TextFormula(res, formula.getDependencies())


    def callZ3(self, data):
        p = Popen([self.z3, "ELIM_QUANTIFIERS=true", "SMTLIB2_COMPLIANT=true", "-smt2", "-in"],
                  stdout=PIPE, stderr=PIPE, stdin=PIPE)
        try:
            res, err = p.communicate(data)
        except KeyboardInterrupt:
            p.kill()
            sys.stdout.write('; Interrupted by Ctrl-C\n')
            exit(1)
            return
        if err:
            raise Exception("ERROR: %s" % err)
        else:
            return res


    pass



class MsatQuantifierElimination(object):
    
    def __init__(self, msat_path, method='fm', tmpDir='/tmp'):
        self.qelim = msat_path
        self.method = method
        self.tmpDir = tmpDir
        return


    def getFreshFilename(self, template="qelim%d.smt2"):
        smt2file = None
        count = random.randint(0, 99999999)
        while smt2file is None or os.path.exists(smt2file):
            smt2file = os.path.join(self.tmpDir, template % count)
            count = count + 1
        return smt2file


    def getFormula(self, smt2):
        arr = smt2.split('\n')
        res = []
        begin = False
        for l in arr:
            if begin:
                if l.split() != []:
                    assert not l.startswith("(assert")
                    res.append(l)
            elif l.startswith("(assert"):
                begin = True
        
        res.pop()

        return self.fixSyntax(" ".join(res))


    def fixSyntax(self, string):
        res = re.sub(r'\(to_real \(- (\d+\.?\d*)\)\)', r'(- \1)', string)
        res = re.sub(r'\(to_real (\d+\.?\d*)\)', r'\1', res)
        return res


    def removeQuantifiers(self, formula):
        tf = ""
        try:
            smt2file = self.getFreshFilename()
            f = open(smt2file, "wt")

            assert formula.nodeType == FNode.EXISTS

            f.write("(set-option :print-success false)\n")
            f.write("(set-logic LRA)\n")
            f.write("\n")
        
            for v in formula.formula.getDependencies():
                f.write("(declare-fun %s %s)\n" % (v.name, v.typename.toString()))

            f.write("\n")

            f.write("(assert \n")
            p = SmtPrinter(f, strictSyntax=True, smt2version=True)
            p.printer(formula.formula)
            f.write(")\n")

            f.close()


            varfile = self.getFreshFilename("varfile%d.vars")
            f = open(varfile, "wt")
        
            for v in formula.variables:
                f.write("%s\n" % (v.name))
            f.close()


            output = self.callElim(smt2file, varfile)
            tf = self.getFormula(common.convertString(output))

            os.remove(smt2file)
            os.remove(varfile)

        except:
            if os.path.exists(smt2file):
                os.remove(smt2file)
            if os.path.exists(varfile):   
                os.remove(varfile)
            raise

        assert(tf is not None)
        return TextFormula(tf, formula.getDependencies())


    def callElim(self, formulafile, varfile):       
        p = Popen([self.qelim, ("-algo=%s" % self.method), "-smt2", formulafile, varfile], stdout=PIPE)
        try:
            res, err = p.communicate()
        except KeyboardInterrupt:
            p.kill()
            sys.stdout.write('; Interrupted by Ctrl-C\n')
            exit(1)
            return
        if err:
            raise Exception("ERROR: %s" % err)
        else:
            return res;


    pass
